//---------------------------------------------------------------------
// BassLine\n MuseScore 3 plugin
//---------------------------------------------------------------------

//=============================================================================
//  MuseScore
//  Music Composition & Notation
//
//  Create bass line for diatonic accordion tablature from a MuseScore music score
//
//  Copyright (C) 2020  Jean-Michel Bencetti
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License version 2
//  as published by the Free Software Foundation and appearing in
//  the file LICENCE.GPL
//
//=============================================================================

//--------------------------------------------------------------------------
/* Ce plugin ajoute une ligne Basse-Accord à une tablature obtenue par DiatonicTab sous MuseScore 
 
  Auteur : Jean-Michel Bencetti
  Version courrante : 0.02
  Date : v0.00 : 2020-03-18 : développement initial
         v0.00.01 : avance par mesure
         v0.00.02 : avance elément
         v0.00.03 : ne regarde pas les mesures, mais la longeur du pattern
         v0.00.04 : utilise la voix 2 de la 2ème portée
	    
  Description version v1.0 :
     
  ----------------------------------------------------------------------------  */
import QtQuick 2.2
import MuseScore 3.0
import QtQuick.Controls 1.1
import QtQuick.Controls.Styles 1.3
import QtQuick.Layouts 1.1
import FileIO 3.0


MuseScore {
version: "0.0.4"
description: "Write Bass/Chord line on a diatonic accordion tablature"
menuPath: "Plugins.DiatonicTab.BassLine"
pluginType: "dialog"


property int margin: 10

   width:  380
   height: 160
   
//-----------------------------------------------------
// Set here the language : FR = French, EN = English
   property string lang: "FR"
//-----------------------------------------------------


//-----------------------------------------------------
// Fichiers JSON pour la mémorisation des parametres
//-----------------------------------------------------
   FileIO {
        id: myParameterFile
        source: homePath() + "/BassLine.json"
        onError: console.log(msg)
        } 
//-----------------------------------------------------
// Critères à choisir dans la boîte de dialogue
//-----------------------------------------------------
  property var parametres: {
        "numPattern" : 0,          // numéro du pattern
        "thePattern" : "",         // Pattern s'il provient de la saisie
        "anacrouse" : "OUI",       // Il y a une anacrouse dans la sélection
        "AccordsComplets" : "OUI", // In affiche les accords complets ou simplifiés
  }

// -------------------------------------------------------------------
// Description de la fenêtre de dialogue
//--------------------------------------------------------------------
 GridLayout {
      id: 'mainLayout'
      anchors.fill: parent
      anchors.margins: 10
      columns: 2

Label {
            width: parent.width
            Layout.columnSpan : 2
            wrapMode: Label.Wrap
            horizontalAlignment: Qt.AlignHCenter
		  font.bold: true
            text:  (lang == "FR") ? qsTr("Ligne Basses/Accords pour accordéon diatonique v" + version) : 
                                    qsTr("Bass/Chords line for diatonic accordion v" + version)
      }


     //------------------------------------------------
     // Choix du Pattern Basse/Accord
     //------------------------------------------------
     Label {
          width: parent.width 
          wrapMode: Label.Wrap
          horizontalAlignment: Qt.AlignRight
          text:  (lang == "FR") ? qsTr("Patterns de base : ") : 
                                  qsTr("Basic patterns : ")
     }

     ComboBox {
         id: comboChoixPattern
         editable: false
               model:  [
               // La liste de choix doit être dans le même ordre que le tableau tabmodeleClavier
               // La liste de choix doit être dans le même ordre que le tableau tabmodeleClavier
               { text:  qsTr("2/2 Ba")        },
               { text:  qsTr("2/4 BaBa")      },
               { text:  qsTr("2/4 Ba")        },
               { text:  qsTr("3/4 Baa")       },
               { text:  qsTr("3/4 BaB")       },
               { text:  qsTr("3/4 B-B")       },
               { text:  qsTr("4/4 BaBa")      },
               { text:  qsTr("4/4 Ba-a")      },
               { text:  qsTr("5/4 BaaBa")     },
               { text:  qsTr("5/4 BaBaa")     },
               { text:  qsTr("6/8 B-aB-a")    },
               { text:  qsTr("6/8 BaaBaa")    },
               { text:  qsTr("7/8 BaaBaBa")   },
               { text:  qsTr("7/8 BaBaBaa")   },
               { text:  qsTr("7/8 B-aBaBa")   },
               { text:  qsTr("7/8 B--a-a-")   },
               { text:  qsTr("8/8 BaaBaaBa")  },
               { text:  qsTr("8/8 BaaBaBaa")  },
               { text:  qsTr("9/8 B-aB-aB-a") },
               { text:  qsTr("9/8 BaaBaaBaa") },
               { text:  qsTr("12/8 B-aB-aB-aB-a")},
               ]
        // Récupère le code du modèle de clavier
        onActivated: { parametres["numPattern"] = index
                       parametres["thePattern"] = model[index].text
                       console.log("Pattern : " + model[index].text)
                       inputTextPattern.text = model[index].text.replace(/(\d+)\/(\d+)\s(\w+)/,"$3")
                      }
          
          }
     Label {
          width: parent.width / 2
          wrapMode: Label.Wrap
          horizontalAlignment: Qt.AlignRight
          text:  (lang == "FR") ? qsTr("Pattern personnalisé : ") : 
                                  qsTr("Personalised pattern : ")
      }

        
     TextField {
          id : inputTextPattern
          
          placeholderText : "BaaBaa"
//     inputMask : "A"               // Que des lettres
          onEditingFinished : {
               console.log("inputTextPattern = " + text)
               parametres["thePattern"] = text
          }
     }
     CheckBox {
          id: cbAnacrouse
          Layout.columnSpan: 1
          text: (lang == "FR" ) ? qsTr("Il y a une anacrouse") :
                                  qsTr("Anacrouse on first measure")
          checked: parametres.anacrouse == "OUI"
                                  
     }
     CheckBox {
          id: cbAccordsComplets
          Layout.columnSpan: 1
          text: (lang == "FR" ) ? qsTr("Accords complets") :
                                  qsTr("Full chords")
          checked: parametres.AccordsComplets == "OUI"
                                  
     }
        //-----------------------------------------------
     Button {
         id: okButton
         isDefault: true
         Layout.columnSpan: 1
         text: qsTr("OK")
         onClicked: {
             // Mémorise les parametres pour la prochaine fois
             parametres["thePattern"] = inputTextPattern.text
             curScore.startCmd();
             // Here start the job --------------
             doBassLine();
             // ---------------------------------
            curScore.endCmd();
            myParameterFile.write(JSON.stringify(parametres))
            Qt.quit();
         }
      } 
Button {
         id: cancelButton
         Layout.columnSpan: 1
         text: (lang=="FR")?qsTr( "Annuler"):qsTr("Cancel")
         onClicked: {
            Qt.quit();
         }
      }
}
// ---------------------------------------------------------------------
// Fonction doBassLine
// Cette fonction fait l'essentiel du travail, elle est appellée dans onRun
//----------------------------------------------------------------------
function doBassLine() {
      var myScore = curScore,                  // Partition en cours
          cursor = myScore.newCursor(),        // Fabrique un curseur pour se déplacer dans les mesures
          endTick,                             // Numéro du dernier élément de la partition ou de la sélection
          staff = 0,                           // Numéro de la portée en cours de traitement
          fullScore = false,                   // Partition entière ou sélection
          accordMg = "",                       // Accord à décomposer en Basse-Accord
          oldAccordMg = "",                    // Accord précédent
          basseAJouer = "" ,                   // Basse à jouer
          accordAJouer = "";                   // Accord à jouer
          
//     console.log("Entrée dans la fonction doBasseLine")

      // Cherche les portées, on ne travaillera pas sur la dernière portée (en général clé de Fa, Basses et Accords)
      var nbPortees = myScore.nstaves
//      console.log("Nombre de portées =",nbPortees)  
      staff = myScore.nstaves - 1
      staff = 1                                       // La ligne de tablature est une portée qui comprend un silence par temps
      cursor.voice    = 1                             // on ne traite que la voix 2 qui doit contenir des silences
      cursor.staffIdx = staff                         // on traite la dernière portée en clé de Fa 

      
      // -----------------------------------------------------------------------
      // Cherche à savoir s'il s'agit de la partition entière ou d'une sélection
      cursor.rewind(Cursor.SELECTION_START)           // rembobine au début de la sélection
      if (!cursor.segment) {                          // pas de sélection
          fullScore = true;
      } else {
          cursor.rewind(Cursor.SELECTION_END)       // passe derrière le dernier segment et fixe tick = 0
          if (cursor.tick === 0) {                  // ceci survient lorsque la sélection contient la dernière mesure
               endTick = curScore.lastSegment.tick + 1;
          } else {
               endTick = cursor.tick;
          }
      }
      // -----------------------------------------------------------------------
      if (fullScore) {                             // si pas de sélection
          cursor.rewind(Cursor.SCORE_START)       // rembobine au début de la partition
      } else {                                     // si sélection
          cursor.rewind(Cursor.SELECTION_START);  // rembobine au début de la sélection
      }


//     console.log("Debut de boucle pour la derniere portée numéro : " + (staff + 1))
      
     var thePattern = inputTextPattern.text 
//     console.log("Pattern choisit dans doBassLine(): " + thePattern)
     
     parametres.anacrouse = (cbAnacrouse.checkedState) ? "OUI" : "NON"
     parametres.AccordsComplets = (cbAccordsComplets.checkedState) ? "OUI" : "NON"
//     console.log("cbAnacrouse = " + cbAnacrouse.checkedState)
     
     cursor.voice = 1
     // S'il y a une Anacrouse, on avance à la première "vraie" mesure
     if (parametres.anacrouse == "OUI" ) cursor.nextMeasure()
     
     // ----------------------------------------------------------------------
     // Boucle pour chaque mesure de la portée ou de la sélection en cours
     // ----------------------------------------------------------------------
     var temps = 0

     while (cursor.measure && (fullScore || cursor.tick < endTick))  {

               // Recherche des accords main gauche (genre Am ou Em ou E7 ...)
               var aCount = 0;
               if (cursor.segment) {
               var annotation = cursor.segment.annotations[aCount];
               while (annotation) {
                    if (annotation.type == Element.HARMONY){                               
//                         console.log("Symbole d'accord : " + annotation.text);
                         accordMg = annotation.text.toUpperCase()
                    }
                           annotation = cursor.segment.annotations[++aCount];     
               }
               }


               // Si on a trouvé un nouvel accord, on cherche la basse et on l'ajoute
               if (accordMg != oldAccordMg) {
                    // Cherche si renversement
                    if (accordMg.match("/")) {
                         accordAJouer = accordMg.split("/")[0]
                         basseAJouer = accordMg.split("/")[1]
                         if (parametres.AccordsComplets != "OUI" ) 
                              accordAJouer = (accordAJouer.match("^[A-Ga-g]#")) ? accordAJouer[0]+"#" : 
                                             (accordAJouer.match("^[A-Ga-g][Bb]")) ? accordAJouer[0] + "b" :
                                             accordAJouer
                         if (accordAJouer) accordAJouer = accordAJouer.toLowerCase()
                    } else {
                         basseAJouer = (accordMg.match("^[A-Ga-g]#")) ? accordMg[0]+"#" : 
                                       (accordMg.match("^[A-Ga-g][Bb]")) ? accordMg[0] + "b" :
                                       accordMg[0].toUpperCase()
                         accordAJouer = (parametres.AccordsComplets == "OUI" ) ? accordMg.toLowerCase()  : basseAJouer.toLowerCase()
                    }
                    oldAccordMg = accordMg
//                    console.log("Basse : " + basseAJouer + " Accord : " + accordAJouer)
               }
                   
               // Fabrication de la ligne Basses/Accords
               var basse = newElement(Element.LYRICS)
               basse.verse = 2                    // 3ème ligne des paroles
               basse.voice = 1                    // 2ème voix
               basse.autoplace = false            // Se place où il faut
               basse.offsetY =  -10               // Se déplace un peu vers le bas (compatibilité avec DiatonicTab
                         
               // Selon le choix "BaBa" ou "Baa" ou "B-aB-a" ou etc ...
//               console.log("Basse ou Accord : " + thePattern[temps] + " Temps : " + temps)
               switch (thePattern[temps]) {
                         case "B" :
                              basse.text = basseAJouer
                              break
                         case "a":
                              basse.text = accordAJouer
                              break
                         default:
                              basse.text = thePattern[temps]
                              break
               }
//               console.log("A afficher : " + basse.text)         
               cursor.add(basse)
               
               // ----------------------------------------------------------
               cursor.next()               // pour aller à l'élément suivant                         

            temps = (temps+1) % thePattern.length
//            console.log("temps : " + temps + "thePattern.length = " + thePattern.length)
 
               
     } // fin du while cursor.smeasure et (fullScore || cursor.tick < endTick)



} // end function doBassLine()
//----------------------------------------------------------------------

  onRun: {
     console.log("Kenavo")
     
     //------------------------------------------------------------------------------
     // Lecture du fichier de parametres
     //------------------------------------------------------------------------------
     console.log("Lecture du fichier de parametres")       
     parametres = JSON.parse(myParameterFile.read()) 
     console.log(JSON.stringify(parametres))
//     console.log("numéro de Pattern choisit : " + comboChoixPattern.model[parametres["numPattern"]].text)
//     console.log("Pattern choisit : " + parametres["thePattern"])
     inputTextPattern.text = parametres["thePattern"]
     //------------------------------------------------------------------------------

     //------------------------------------------------------------------------------
     //  Remet le bon pattern dans la comboBox
     //------------------------------------------------------------------------------
     comboChoixPattern.currentIndex = parametres["numPattern"]

  } // end onRun

} // MuseScore
